// test-Benaphore
//
// Copyright © 1996-1998 Arcane Dragon Software
//                       Chris Herborth (chrish@kagi.com)
//
// Test the ChBenaphore object by throwing around a few threads.
//
// As an exercise for the reader, you could modify this to spawn n threads
// based on a command-line option...

#include <kernel/OS.h>
#include <stdio.h>

#include "Ch/ChBenaphore.h"

static char *_rcs_id = "$Id: test-Benaphore.cpp,v 1.2 1998/04/14 00:34:05 chrish Exp $";

// Prototypes.
long thread1( void *data );
long thread2( void *data );
long thread3( void *data );

// A global to smooth access to stdout.
ChBenaphore *stdout_guard;

// Print to stdout, one at a time.
static void safe_puts( const char *str )
{
	status_t retval;
	retval = stdout_guard->Lock();
	if( retval == B_OK ) {
		puts( str );
	}
	stdout_guard->Unlock();
}

// Thread 1
long thread1( void *data )
{
	status_t retval;

	ChBenaphore *mutex = (ChBenaphore *)data;

	while( 1 ) {
		safe_puts( "--- 1 - Trying to lock.        ---" );
		retval = mutex->Lock();
		safe_puts( "--- 1 - Mutex locked.          ---" );
		safe_puts( "--- 1 - sleeping for 4 seconds ---" );
		snooze( 1000.0 * 1000.0 * 4 );
		safe_puts( "--- 1 - done sleeping          ---" );
		safe_puts( "--- 1 - Unlocking.             ---" );
		mutex->Unlock();
		safe_puts( "--- 1 - Going around again     ---" );
	}
}

// Thread 2
long thread2( void *data )
{
	status_t retval;

	ChBenaphore *mutex = (ChBenaphore *)data;

	while( 1 ) {
		safe_puts( "*** 2 - Locking.               ***" );
		retval = mutex->Lock();
		safe_puts( "*** 2 - Locked!                ***" );
		safe_puts( "*** 2 - sleeping for 3 seconds ***" );
		snooze( 1000.0 * 1000.0 * 3 );
		safe_puts( "*** 2 - done sleeping          ***" );
		safe_puts( "*** 2 - Unlocking.             ***" );
		mutex->Unlock();
	}
}

// Thread 3
long thread3( void *data )
{
	status_t retval;

	ChBenaphore *mutex = (ChBenaphore *)data;

	while( 1 ) {
		safe_puts( "!!! 3 - Locking.               !!!" );
		retval = mutex->Lock();
		safe_puts( "!!! 3 - snoozing for 9 seconds !!!" );
		snooze( 1000.0 * 1000.0 * 9 );
		safe_puts( "!!! 3 - Unlocking.             !!!" );
		mutex->Unlock();
	}
}

// main
int main( int argc, char **argv )
{
#pragma unused (argc,argv)
	printf( "ChBenaphore test\n" );
	printf( "%s\n", _rcs_id );
	printf( "\n" );

	stdout_guard = new ChBenaphore( "stdout guard" );
	if( !stdout_guard->IsValid() ) {
		fprintf( stderr, "Can't create Benaphore to guard stdout...\n" );
		return -1;
	}

	ChBenaphore the_mutex( "main's mutex" );

	// Now we create some threads and let them fight over the Benaphore.
	thread_id t1 = spawn_thread( thread1,
								 "Thread 1",
								 B_NORMAL_PRIORITY,
								 (void *)&the_mutex );
	thread_id t2 = spawn_thread( thread2,
								 "Thread 2",
								 B_NORMAL_PRIORITY,
								 (void *)&the_mutex );
	thread_id t3 = spawn_thread( thread3,
								 "Thread 3",
								 B_NORMAL_PRIORITY,
								 (void *)&the_mutex );
	if( t1 < B_NO_ERROR || t2 < B_NO_ERROR || t3 < B_NO_ERROR ) {
		fprintf( stderr, "main(): Argh, couldn't spawn our test threads.\n" );
		return -1;
	}

	long retval = resume_thread( t1 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Couldn't start thread 1.\n" );
		return -1;
	}
	retval = resume_thread( t2 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Couldn't start thread 2.\n" );
		return -1;
	}
	retval = resume_thread( t3 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Couldn't start thread 3.\n" );
		return -1;
	}

	// Let 'em run for 60 seconds.
	snooze( 60.0 * 1000.0 * 1000.0 );

	// Kill 'em all after grabbing the Benaphore to make sure nobody has it
	// locked when they vanish.  Probably not necessary, but I'll bet it's
	// good form.
	safe_puts( "main(): Locking.\n" );
	status_t mutexretval = the_mutex.Lock();
	safe_puts( "main(): Got the lock, killing threads.\n" );
	retval = kill_thread( t1 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Error (%d) killing thread 1!\n", retval );
	}
	retval = kill_thread( t2 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Error (%d) killing thread 2!\n", retval );
	}
	retval = kill_thread( t3 );
	if( retval != B_NO_ERROR ) {
		fprintf( stderr, "main(): Error (%d) killing thread 3!\n", retval );
	}

	// Clean up.
	mutexretval = stdout_guard->Lock();
	delete stdout_guard;

	// Bye bye now.
	return 0;
}
